local _G = _G

-- addon name and namespace
local ADDON, NS = ...

local Addon = LibStub("AceAddon-3.0"):GetAddon(ADDON)

-- get translations
local L = LibStub:GetLibrary("AceLocale-3.0"):GetLocale(ADDON)

-- lib containing map data
local LibMapData = LibStub:GetLibrary("HereBeDragons-1.0")

-- LibTourist
local LibTourist = LibStub:GetLibrary("LibTourist-3.0")

-- the module name
local MODULE      = "Bloodhound"
local DESCRIPTION = L["Calculation of force like in Bloodhound."]

-- local functions
local pairs  = pairs
local pow    = math.pow
local abs    = math.abs
local atan2  = math.atan2
local floor  = math.floor
local sin    = math.sin
local cos    = math.cos
local sqrt   = math.sqrt

local PI     = math.pi
local HALFPI = PI / 2
local RAD    = 180 / PI

-- aux variables
local _

-- the module
local Algorithm = NS:CreateAlgorithm(MODULE, DESCRIPTION)

Algorithm:SetReady(LibMapData and LibTourist and true or false)

local MIN_WINDOW = 1
local MAX_WINDOW = 90

-- aux vars force calculation
local forces = {}

do
	for i = -179, 180 do 
		forces[i] = {} 
	end
	
	-- take care of overflow buckets
	for i = -179-MAX_WINDOW, -180 do 
		forces[i] = forces[i+360]
	end
end

local defaults = {
	age = 2,
	distance = 2,
}

local database = nil

local options = {}

local degrees = {
	[0] = L["Constant"],
	[1] = L["Linear"],
	[2] = L["Quadratic"],
	[3] = L["Cubic"],
}

-- register
NS:RegisterModule(Algorithm)

function Algorithm:CalculateForce(inlet, pX, pY, zoneName, zoneMap)
	if not NS:IsValidInlet(inlet) or type(pX) ~= "number" or type(pY) ~= "number" then
		return 0, 0
	end
	
	-- reset data
	for angle, data in pairs(forces) do
		data.x = 0
		data.y = 0
		data.count = 0
		data.magnitude = 0 -- squared
	end
	
	local now = time()
	
	-- convert to continent position - we are in a normal zone so floor is always 0
	local pgX, pgY = LibMapData:GetWorldCoordinatesFromZone(pX, pY, zoneMap, 0)

	local minAge           = Addon:GetSetting("minNodeAge")
	local maxAge           = Addon:GetSetting("maxNodeAge")
	local visitingDistance = Addon:GetSetting("visitingDistance")	
	
	local exponentAge      = self:GetSetting("age")
	local exponentDistance = self:GetSetting("distance")
	
	minAge = maxAge < minAge and maxAge or minAge
	
	-- makes sure that the force of a node at max age in exactly visiting range is 1
	local normalization = (visitingDistance^exponentDistance) / (maxAge^exponentAge)
	
	local totalCount   = 0
	local visitedCount = 0
	
	local forceX, forceY = 0, 0
	
	for profession, info in Addon:IterateTrackingSkills() do
		if info.active then
			for zoneId, zone in Addon:IterateSearchZones() do
				local map = zone == zoneName and zoneMap or LibTourist:GetTexture(zone)
				
				if map then				
					for state, encLoc, nodeName in inlet:GetNodesIterator(zoneId, profession) do
						local nX, nY = NS:DecodeLoc(encLoc)
												
						-- convert to continent position - we are in a normal zone so floor is always 0
						local ngX, ngY = LibMapData:GetWorldCoordinatesFromZone(nX, nY, map, 0)
						
						local dX = ngX - pgX
						local dY = ngY - pgY
						local distance = sqrt(dX * dX + dY * dY)
						
						totalCount = totalCount + 1

						-- update visiting time
						if distance < visitingDistance then
							Addon:SetVisited(zone, encLoc, now)

							visitedCount = visitedCount + 1
						else
							local age = Addon:GetAge(zone, encLoc, now, maxAge)
							
							if age < maxAge then
								visitedCount = visitedCount + 1
							end
							
							if age >= minAge then
								--[=====[								
								-- for exponents == 1
								local force = age/distance
								local angle = math.atan2(dX, dY)
								local fX = force * math.cos(angle) = age/distance * dX/distance = age/distance^2 * dX
								local fY = force * math.sin(angle) = age/distance * dY/distance = age/distance^2 * dY
								--]=====]
								
								local forceD = normalization * age^exponentAge / distance^(exponentDistance+1)
								
								forceX = forceX + forceD * dX
								forceY = forceY + forceD * dY
							end
						end
					end -- for node
				end					
			end -- for zone
		end -- if active
	end -- for profession
	
	local magnitude = sqrt(forceX*forceX + forceY*forceY)	
	local direction = atan2(forceY, -forceX)
	
	-- adjusted by pi/2 to adjust to orientation where north is at 0
	direction = direction - HALFPI
	
	-- this represents the distance if magnitude was represented by a single node (of max age)
	local singularDistance = magnitude == 0 and visitingDistance*16 or ((visitingDistance^exponentDistance) / magnitude)^(1/exponentDistance)

	-- rating is suppossed to lie between 1 (lowest) and 5 (highest)
	-- the addon will take care of checking these boundaries
	-- 5 rating if distance is less then visitingDistance*2^2, 3 if visitingDistance*2^3
	-- NOTE: the rating of this algorithm is a bit difficult since opposite nodes reduce the resulting force
	local rating = 6 - log(singularDistance/visitingDistance)/LN2

	return magnitude, direction, rating, totalCount, visitedCount
end

function Algorithm:SetupOptions(db)
	if type(db) ~= "table" then		
		return {}
	end

	database = db
	
	for key, value in pairs(defaults) do
		if not database[key] then
			database[key] = value
		end
	end
	
	NS:ClearTable(options)

	options.type = 'group'
	options.name = self:GetName()
	options.desc = L["Settings for "] .. self:GetName()
	options.args = {}
	
	-- description
	options.args["description"] = {
		type = "description",
		name = NS:Colorize("Yellow", self:GetName()) .. ": " .. self:GetDescription(),
		order = 1,
	}	

	-- age
	options.args["age"] = {
		type = 'select',
		name = L["Age"],
		desc = L["Weight of age."],
		order = 2,
		get = function()
			return self:GetSetting("age")
		end,	
		set = function(info, degree)
			self:SetSetting("age", degree)
		end,
		values = degrees,
	}
	
	-- distance
	options.args["distance"] = {
		type = 'select',
		name = L["Distance"],
		desc = L["Weight of distance."],
		order = 3,
		get = function()
			return self:GetSetting("distance")
		end,	
		set = function(info, degree)
			self:SetSetting("distance", degree)
		end,
		values = degrees,
	}
	
	return options
end

function Algorithm:GetSetting(setting)
	return database and database[setting] or default[setting]
end

function Algorithm:SetSetting(setting, value)
	if database then
		database[setting] = value
	end
end

-- test
function Algorithm:Debug(msg)
	Addon:Debug("(Algorithm:" .. MODULE .. ") " .. msg)
end
