--[[
Copyright (c) 2015, Bernhard Voigt <tritium@noisywires.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright owner nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--]]
--[[
	Library that provides centralized handling for re-parenting the minimap.
--]]
local MAJOR = "LibHijackMinimap-1.0"
local MINOR = 1000 + tonumber(("$Revision: 1 $"):match("%d+"))
assert(LibStub, MAJOR.." requires LibStub")

local lib = LibStub:NewLibrary(MAJOR, MINOR)

if not lib then 
	return 
end

lib.callbacks = lib.callbacks or LibStub("CallbackHandler-1.0"):New(lib)
local callbacks = lib.callbacks

local _G = _G

local Minimap = Minimap

local _

local tablestack = setmetatable({}, {__mode = 'k'})

function ClearTable(t)
	if type(t) ~= "table" then
		return
	end
	
	for k, v in pairs(t) do
		t[k] = nil
	end
end

---
--- external functions/constants
---
lib.external = {
	'RegisterHijacker',
	'UnregisterHijacker',
	'IsHijackerRegistered',
	'RegisterClient',
	'UnregisterClient',
	'IsClientRegistered',
	'IterateHijackers',
	'IterateClients',
	'IterateClientsForHijacker',
	'GetClientSettingsActive',
	'SetActive',
	'IsActive',
	'Blacklist',
	'IsBlacklisted',
	'HijackMinimap',
	'ReleaseMinimap',
	'IsMinimapReplaced',
}

lib.hijackers = {}
lib.clients = {}

lib.hijackerLookupByToken = {}

lib.optionsToken = nil
lib.active = true

function lib.Embed(_, handler)
	local self, usage = lib, 'Embed(handler)'

	self:CheckArgument(usage, 'handler', 'table', handler)

	for _, name in pairs(self.external) do
		handler[name] = self[name]
	end -- do
	
	self['embeds'][handler] = true

	return handler
end

function lib.Library(_)
	local self = lib

	return self:Embed({})
end

function lib:CheckArgument(func, name, argtype, arg, defarg)
	if arg == nil and defarg ~= nil then
		return defarg
	elseif type(arg) == argtype then
		return arg
	else
		error(string.format("%s: '%s' - %s%s expected got %s", func, name, (defarg ~= nil) and 'nil or ' or '', argtype, type(arg)), 3)
	end
end

--- Registers hijacker for executing minimap replacements.
--
-- @param name Name of the hijacker.
-- @param token The private token of the hijacker. Used so third parties cannot act instead of the original hijacker.
function lib:RegisterHijacker(name, token)
	if type(name) ~= "string" then
		error(("Usage: RegisterHijacker(name, token): 'name' - Expected string but got '%s'."):format(type(name)), 2)	
	end
	
	if self.hijackers[name] then
		error(("Usage: RegisterHijacker(name, token): 'name' - Hijacker with name '%s' already exists."):format(type(name)), 2)	
	end

	-- generate token if not provided
	token = token or {}
	
	if self.hijackerLookupByToken[token] then
		error(("Usage: RegisterHijacker(name, token): 'token' - Token '%s' already in use."):format(type(token)), 2)	
	end
	
	local hijacker = next(tablestack) or {}
	tablestack[hijacker] = nil
	
	-- assign unique token to registered hijacker that is required to unregister, so that not just anyone can unregister hijackers via name 
	hijacker.token = token
	hijacker.name = name
	hijacker.blacklist = next(tablestack) or {}
	tablestack[hijacker.blacklist] = nil
	
	self.hijackers[name] = hijacker
	self.hijackerLookupByToken[token] = name
	
	self.callbacks:Fire("LHM_HIJACKER_REGISTERED", name)
	
	return hijacker.token
end

--- Unregisters client for handling of minimap replacements.
--
-- @param token The private token of the hijacker registered with.
function lib:UnregisterHijacker(token)
	local name = self.hijackerLookupByToken[token]

	if not name then
		return
	end

	local hijacker = self.hijackers[name]

	if hijacker then
		if self:IsMinimapReplaced(name) then
			self:ReleaseMinimap(token)			
		end
	
		self.hijackers[name] = nil 

		ClearTable(hijacker.blacklist)
		tablestack[hijacker.blacklist] = true
		
		ClearTable(hijacker)
		tablestack[hijacker] = true
		
		self.callbacks:Fire("LHM_HIJACKER_UNREGISTERED", name)
	end
end

--- Returns true if hijacker with given name is registered.
--
-- @param name Name of the hijacker to check.
function lib:IsHijackerRegistered(name)
	return self.hijackers[name] and true
end

--- Registers client for handling of minimap replacements.
--
-- @param name Name of the client.
-- @param handle Handle of the client. If arg is string global variable of that name will be looked up and used.
-- @param func Function handling the minimap replacement for client. If arg is string function of that name will be looked up in handle.
function lib:RegisterClient(name, handle, func)
	if type(name) ~= "string" then
		error(("Usage: RegisterClient(name, handle, func): 'name' - Expected string but got '%s'."):format(type(name)), 2)	
	end
	
	if self.clients[name] then
		error(("Usage: RegisterClient(name, handle, func): 'name' - Client with name '%s' already exists."):format(type(name)), 2)	
	end
	
	if type(handle) == "string" then
		handle = _G[handle]
	end

	if type(handle) ~= "table" then
		error(("Usage: RegisterClient(name, handle, func): 'handle' - Expected table but got '%s'."):format(type(handle)), 2)	
	end
		
	if type(func) == "string" then
		if not handle[func] then
			error(("Usage: RegisterClient(name, handle, func): 'func' - Could not find function for handle %s with name '%s'."):format(tostring(handle), tostring(func)), 2)	
		end
	
		func = handle[func]
	end

	if type(func) ~= "function" then
		error(("Usage: RegisterClient(name, handle, func): 'func' - Expected function but got '%s'."):format(tostring(func)), 2)	
	end

	local client = next(tablestack) or {}
	tablestack[client] = nil

	client.name = name
	client.handle = handle	
	client.func = func
	
	self.clients[name] = client
	
	self.callbacks:Fire("LHM_CLIENT_REGISTERED", name)
	
	self:TryHijackClient(name)
	
	return true	
end

--- Unregisters client for handling of minimap replacements.
--
-- @param handle Handle the client registered with.
function lib:UnregisterClient(handle)
	for name, client in pairs(self.clients) do
		if client.handle == handle then
			if client.hijacker then
				self:HijackClient(name, Minimap, client.hijacker)
			end
		
			self.clients[name] = nil 
			
			ClearTable(client)
			tablestack[client] = true
			
			self.callbacks:Fire("LHM_CLIENT_UNREGISTERED", name)
			
			return
		end
	end	
end

--- Returns true if client with given name is registered.
--
-- @param name Name of the client to check.
function lib:IsClientRegistered(name)
	return self.clients[name] and true	
end

-- iterators:
-- IterateHijackers
-- IterateClients
-- IterateClientsForHijacker
do --Do-end block for iterators
	local emptyTbl = {}
	local tablestack = setmetatable({}, {__mode = 'k'})

	local function KeyOnlyIter(t, prestate)
		if not t then 
			return nil 
		end

		if t.iterator then
			local key = t.iterator(t.t, prestate)

			if key then
				return key
			end				
		end
		
		tablestack[t] = true
		return nil, nil		
	end

	local function IterateKeys(toIterate)
		local tbl = next(tablestack) or {}		
		tablestack[tbl] = nil
		
		local data = toIterate or emptyTbl
		
		local iterator, t, state = pairs(data)
		
		tbl.iterator   = iterator
		tbl.t          = t
		
		return KeyOnlyIter, tbl, state
	end
	
	--- Returns iterator that provides all registered hijackers.
	--
	function lib:IterateHijackers()
		return IterateKeys(self.hijackers)
	end

	--- Returns iterator that provides all registered clients.
	--
	function lib:IterateClients()
		return IterateKeys(self.clients)
	end
	
	local function ClientsForHijackerIter(t, prestate)
		if not t then 
			return nil 
		end

		if t.iterator and t.hijacker then
			local name, client = t.iterator(t.t, prestate)
			while client do
				if not t.hijacker.blacklist[name] then
					return name
				end

				name, client = t.iterator(t.t, name)
			end
		end
		
		tablestack[t] = true
		return nil, nil		
	end	
	
	--- Returns iterator that provides the client names that are not blacklisted with a hijacker.
	--
	-- @param hijackerName Name of hijacker for which the iterator should provide the client names.
	function lib:IterateClientsForHijacker(hijackerName)
		local tbl = next(tablestack) or {}		
		tablestack[tbl] = nil
		
		local data = self.clients or emptyTbl
		
		local iterator, t, state = pairs(data)
		
		tbl.iterator   = iterator
		tbl.t          = t
		tbl.hijacker   = self.hijackers[hijackerName]
		
		return ClientsForHijackerIter, tbl, state
	end
end

--- Returns true if any instance registered to apply settings.
--
function lib:GetClientSettingsActive()
	return self.optionsToken and true or false
end

--- Checks and assigns token for instance setting the options. Produces error if different token is already set.
--
-- **Note:** This function is for internal use.
-- @param optionsToken The private token of the instance doing the setting. Note: There can only be a single instance doing all the settings. (Possibly LibHijackMinimap-1.0-Options)
-- @param msg Message used in error.
function lib:CheckOptionsToken(optionsToken, msg)
	if not msg then
		msg = "CheckOptionsToken(optionsToken, msg)"
	end
	
	if not optionsToken then
		error(("Usage: %s: 'optionsToken' - Expected value but got '%s'."):format(tostring(msg), tostring(optionsToken), 2))
	end
	
	if not self.optionsToken then
		self.optionsToken = optionsToken
	end

	if optionsToken ~= self.optionsToken then	
		error(("Usage: %s: 'optionsToken' - Token '%s' is different from registered token '%s'. Only a single instance may apply the settings."):format(tostring(msg), tostring(optionsToken), tostring(self.optionsToken)), 2)
	end
end


--- Activates or deactivates handling of minimap replacements.
--  If deactivated all replacements are reverted and hijackers registering replacements are registered but not executed.
--  If activated all replacements registered for hijackers are forwarded to the respective clients and new replacements are handled again.
--
-- @param active flag indicating if operation should be active or not.
-- @param optionsToken The private token of the instance doing the setting. Note: There can only be a single instance doing all the settings. (Possibly LibHijackMinimap-1.0-Options)
function lib:SetActive(active, optionsToken)
	self:CheckOptionsToken(optionsToken, "SetActive(active, optionsToken)")
	
	active = active and true or false
	
	if active == self.active then
		return
	end

	self.active = active
	
	self.callbacks:Fire("LHM_ACTIVE", active)
	
	-- adjust hijacking	
	for clientName in self:IterateClients() do		
		if active then
			-- hijack client
			self:TryHijackClient(clientName)
		else
			-- release client
			local client = self.clients[clientName]
		
			if client.hijacker then		
				self:HijackClient(clientName, Minimap, client.hijacker)
			end
		end		
	end
end

--- Returns true if handling of minimap replacement is activated. 
--  On start up the handling is active.
function lib:IsActive()
	return self.active
end

--- Blacklist client for hijacker for minimap replacements.
--
-- @param hijackerName Name of hijacker to blacklist the clients for.
-- @param clientNames Name of client to blacklist for hijacker.
-- @param enable Boolean indicating if client should be blacklisted or not.
-- @param optionsToken The private token of the instance doing the setting. Note: There can only be a single instance doing all the settings. (Possibly LibHijackMinimap-1.0-Options)
function lib:Blacklist(hijackerName, clientName, enable, optionsToken)
	self:CheckOptionsToken(optionsToken, "BlacklistClientsForHijacker(hijackerName, clientNames, optionsToken)")
	
	if hijackerName == nil or not self.hijackers[hijackerName] then
		return
	end
	
	if clientName == nil or not self.clients[clientName] then
		return
	end
	
	enable = enable and true or false
	
	if enable == self:IsBlacklisted(hijackerName, clientName) then
		return
	end
	
	local hijacker = self.hijackers[hijackerName]
	local client = self.clients[clientName]
	
	hijacker.blacklist[clientName] = enable and true or nil

	self.callbacks:Fire("LHM_HIJACKER_BLACKLIST_CHANGED", hijackerName, clientName, enable)
	
	if enable then
		-- remove from current hijacker
		if client.hijacker == hijackerName then
			self:HijackClient(clientName, Minimap, hijackerName)
		end
		
		-- do hijack if any other hijacker with active replacement can handle client
		self:TryHijackClient(clientName)		
	else
		-- hijack client if not hijacked by another
		if self:IsMinimapReplaced(hijackerName) and not client.hijacker then
			self:HijackClient(clientName, self.hijackers[hijackerName].minimap, hijackerName)
		end
	end
end

--- Returns true id the client is blacklisted for hijacker otherwise false.
--
-- @param hijackerName Name of the hijacker for the lookup.
-- @param clientName Name of the client for the lookup.
function lib:IsBlacklisted(hijackerName, clientName)
	return self.hijackers[hijackerName] and self.hijackers[hijackerName].blacklist[clientName] or false
end

--- Notify replacement of minimap with different frame for hijacker with given token.
--
-- @param token The private token of the hijacker. Used so third parties cannot act instead of the original hijacker.
-- @param frame The frame used by hijacker as minimap replacement.
function lib:HijackMinimap(token, frame)
	local hijackerName = self.hijackerLookupByToken[token]

	if not hijackerName then
		error(("Usage: HijackMinimap(token, frame): 'token' - Hijacker token '%s' is unknown."):format(tostring(token)), 2)
	end

	if type(frame) == "string" then
		frame = _G[frame]
	end
	
	-- make sure we have a frame	
	if type(frame) ~= "table" or type(frame.IsObjectType) ~= "function" or not frame:IsObjectType("Frame") or frame == Minimap then
		frame = nil
	end
	
	local hijacker = self.hijackers[hijackerName]

	if hijacker and hijacker.minimap ~= frame then	
		hijacker.minimap = frame

		frame = frame or Minimap		
		
		if not self:IsActive() and frame ~= Minimap then
			return
		end
			
		-- hijack for all clients allowed for hijacker
		for clientName in self:IterateClientsForHijacker(hijackerName) do
			local client = self.clients[clientName]
		
			-- client not yet hijacked or by current hijacker
			if not client.hijacker or client.hijacker == hijackerName then
				self:HijackClient(clientName, frame, hijackerName)
			end
		end

		-- on release move clients to next hijacker with valid replacement
		if self:IsActive() and frame == Minimap then
			for clientName in self:IterateClientsForHijacker(hijackerName) do
				self:TryHijackClient(clientName)
			end
		end
	end
end

--- Notify restoration of minimap for hijacker with given token.
--
-- @param token The private token of the hijacker. Used so third parties cannot act instead of the original hijacker.
function lib:ReleaseMinimap(token)
	self:HijackMinimap(token, Minimap)
end

--- Returns frame used as minimap replacement by hijacker with given name.
--
-- **Note:** This function is for internal use.
-- @param hijackerName Name of hijacker to check.
function lib:GetMinimapReplacement(hijackerName)
	return self.hijackers[hijackerName] and self.hijackers[hijackerName].minimap
end

--- Returns true if minimap is replaced by hijacker with given name.
--
-- @param hijackerName Name of hijacker to check.
function lib:IsMinimapReplaced(hijackerName)
	return self:GetMinimapReplacement(hijackerName) and true or false
end

--- Returns name of current hijacker for client with given name.
--
-- @param clientName Name of client to check.
function lib:GetCurrentHijacker(clientName)
	local client = self.clients[clientName]

	return client and client.hijacker or nil
end

--- Tries to find a valid hijacker for client with active minimap replacement. If found replacement is executed for client.
--
-- **Note:** This function is for internal use.
-- @param clientName Name of client to hijack.
function lib:TryHijackClient(clientName)
	for hijackerName in self:IterateHijackers() do
		if self:IsMinimapReplaced(hijackerName) and not self:IsBlacklisted(hijackerName, clientName) then
			self:HijackClient(clientName, self.hijackers[hijackerName].minimap, hijackerName)
			
			break
		end
	end
end

--- Executes callback handling minimap hijack for client.
--
-- **Note:** This function is for internal use.
-- @param clientName Name of client to execute callback.
-- @param frame The frame to be used as minimap.
-- @param hijackerName Name of the hijacker.
function lib:HijackClient(clientName, frame, hijackerName)
	local client = self.clients[clientName]
	
	if client and frame then
		client.hijacker = frame ~= Minimap and hijackerName or nil
		client.func(client.handle, frame)
	end
end
