local _G = _G

local ADDON, NS = ...

-- setup libs
local LibStub = _G.LibStub

-- get translations
local L   = LibStub:GetLibrary("AceLocale-3.0"):GetLocale(ADDON)

-- lib containing map data
local LibMapData = LibStub:GetLibrary("HereBeDragons-1.0")

-- LibTourist
local LibTourist = LibStub:GetLibrary("LibTourist-3.0")

-- LibNodeTypeInfo
local LibNTI = LibStub:GetLibrary("LibNodeTypeInfo-1.0")

-- local functions
local ipairs   = ipairs 
local pairs    = pairs 
local format   = string.format 
local strlower = strlower
local strfind  = strfind
local strsub   = strsub
local tostring = tostring
local type     = type
local abs      = math.abs
local atan2    = math.atan2
local floor    = math.floor
local sin      = math.sin
local cos      = math.cos
local sqrt     = math.sqrt
local tinsert  = table.insert
local tremove  = table.remove

local GetMapNameByID       = _G.GetMapNameByID
local GetNumTrackingTypes  = _G.GetNumTrackingTypes
local GetProfessionInfo    = _G.GetProfessionInfo
local GetSpellInfo         = _G.GetSpellInfo
local GetTime              = _G.GetTime
local GetTrackingInfo      = _G.GetTrackingInfo
local IsFlying             = _G.IsFlying
local IsResting            = _G.IsResting
local SetTracking          = _G.SetTracking
local UnitAffectingCombat  = _G.UnitAffectingCombat
local UnitLevel            = _G.UnitLevel
local UnitOnTaxi           = _G.UnitOnTaxi
local GetPlots             = _G.C_Garrison.GetPlots
local GetOwnedBuildingInfo = _G.C_Garrison.GetOwnedBuildingInfo

local LE_FOLLOWER_TYPE_GARRISON_6_0 = _G.LE_FOLLOWER_TYPE_GARRISON_6_0

local PI     = math.pi
local TWOPI  = PI * 2

-- aux variables
local _

-- colors
NS.HexColors = {
	Blueish          = "04adcb",
	Brownish         = "eda55f",
	GrayOut          = "808080",
	Green            = "00ff00",
	Orange           = "e54100",
	Red              = "ff0000",
	White            = "ffffff",
	Yellow           = "ffff00",
	[L["Mining"]]    = "9ee3ef",
	[L["Herbalism"]] = "b1ef9e",
	[L["Fishing"]]   = "d8a8ec",
	[L["Logging"]]   = "f1c087",
}

-- node sorting
function NS.compareNodes(a, b)
	local profession = LibNTI:GetProfession(a)
	
	if profession ~= LibNTI:GetProfession(b) then
		return profession < LibNTI:GetProfession(b)
	end

	if LibNTI:GetExpansion(profession, a) < LibNTI:GetExpansion(profession, b) then
		return true
	elseif LibNTI:GetExpansion(profession, a) > LibNTI:GetExpansion(profession, b) then
		return false
	else
		return LibNTI:GetName(profession, a) < LibNTI:GetName(profession, b)
	end
end

-- addon and locals
local Addon = LibStub:GetLibrary("AceAddon-3.0"):NewAddon(ADDON, "AceEvent-3.0", "AceConsole-3.0", "AceBucket-3.0", "AceTimer-3.0")

-- addon constants
Addon.MODNAME   = "Magneto"
Addon.FULLNAME  = "Magneto"
Addon.SHORTNAME = "Magneto"

Addon.PROFESSIONS = { 
	[LibNTI.PROFESSION_MINING] = {
		name       = L["Mining"],
		abbrev     = L["M"], 
	},
	[LibNTI.PROFESSION_HERBALISM] = {
		name       = L["Herbalism"],
		abbrev     = L["H"], 
	}, 
	[LibNTI.PROFESSION_FISHING] = {
		name       = L["Fishing"],
		abbrev     = L["F"], 
	}, 
	[LibNTI.PROFESSION_LOGGING] = {
		name       = L["Logging"],
		abbrev     = L["L"], 
	}, 
}

Addon.mapSkillId = {
	[186] = LibNTI.PROFESSION_MINING,    -- key is skillLine returned by GetProfessionInfo() for mning
	[182] = LibNTI.PROFESSION_HERBALISM, -- key is skillLine returned by GetProfessionInfo() for herbalism
	[356] = LibNTI.PROFESSION_FISHING,   -- key is skillLine returned by GetProfessionInfo() for fishing
}

-- addon states
Addon.INACTIVE = "Inactive"
Addon.IDLE     = "Idle"
Addon.PAUSED   = "Paused"
Addon.TRACKING = "Tracking"

-- level range options
Addon.LEVEL_ALL   = "all"
Addon.LEVEL_RANGE = "range"
Addon.LEVEL_LOW   = "low"

-- multi zone options
Addon.MULTI_NEVER  = "never"
Addon.MULTI_ALWAYS = "always"
Addon.MULTI_FLYING = "flying"

Addon.MIN_NODE_AGE =  600
Addon.MAX_NODE_AGE = 5400

Addon.MIN_NODE_IGNORE_TIME =   0
Addon.MAX_NODE_IGNORE_TIME = 600

local SKILL_UP_PATTERN      = ERR_SKILL_UP_SI:gsub("%%s", "%(%.-%)"):gsub("%%d", "%(%d+%)")
local SKILL_LEARNED_PATTERN = ERR_SKILL_GAINED_S:gsub("%%s", "%(%.-%)")

local ICON          = "Interface\\Addons\\"..ADDON.."\\icon.tga"
local ICON_DISABLED = "Interface\\Addons\\"..ADDON.."\\inactive.tga"

local LUMBERMILL_SMALL  =  40
local LUMBERMILL_MEDIUM =  41
local LUMBERMILL_LARGE  = 138

local lumbermillSize = {
    [LUMBERMILL_SMALL]  = L["Small"],
    [LUMBERMILL_MEDIUM] = L["Medium"],
    [LUMBERMILL_LARGE]  = L["Large"],
}

-- modules
local Options       = nil
local Tooltip       = nil
local MinimapButton = nil
local DataBroker    = nil
local Inlets        = nil
local Outlets       = nil
local Algorithms    = nil

local updateHandler = {
	active              = "UpdateActiveSetting",
	minimap             = "UpdateMinimapSetting",
	inlet               = "UpdateInletSetting",
	outlet              = "UpdateOutletSetting",
	algorithm           = "UpdateAlgorithmSetting",
	smoothness          = "UpdateSmoothnessSetting",
	pauseWhenResting    = "UpdateStateSetting",
	pauseInCombat       = "UpdateStateSetting",
	level               = "UpdateZoneLevelSetting",
	multi               = "UpdateMultiZoneSetting",
	hostile             = "UpdateZoneSetting",
	contested           = "UpdateZoneSetting",
	sanctuary           = "UpdateZoneSetting",
	city                = "UpdateZoneSetting",
	pvp                 = "UpdateZoneSetting",
	zoneSpecific        = "UpdateNodeSetting",
	expansions          = "UpdateNodeSetting",
	blacklist           = "UpdateNodeSetting",
}

-- infrastructure
function Addon:OnInitialize()
	-- set module references
	Options       = self:GetModule("Options")
	Tooltip       = self:GetModule("Tooltip")
	MinimapButton = self:GetModule("MinimapButton")
	DataBroker    = self:GetModule("DataBroker")
	Inlets        = self:GetModule("Inlets")
	Outlets       = self:GetModule("Outlets")
	Algorithms    = self:GetModule("Algorithms")

	-- debug
	self.debug = false
	
	-- watch frame
	self.frame = CreateFrame("Frame")

	-- player level
	self.playerLvl = UnitLevel("player")
	
	-- state 
	self.state = self.INACTIVE
	
	-- running state
	self.running = false

	-- the force
	self.forceMagnitude    = 0
	self.forceDirection    = 0
	self.forceSinDir       = 0
	self.forceCosDir       = 1
	self.forceRating       = 0
	self.forceTotalCount   = 0
	self.forceVisitedCount = 0
	
	-- smooth force
	self.smoothMagnitude = 0
	self.smoothDirection = 0
	self.smoothSinDir    = 0
	self.smoothCosDir    = 1
	
	self.smoothStep      = PI

	-- profession skills
	self.skill = {}
	
	-- tracking states
	self.tracking = {}
	
	for profession in pairs(Addon.PROFESSIONS) do
		self.skill[profession] = 0
	
		self.tracking[profession] = {
			name    = nil,
			spellId = nil,
			ready   = false,
			active  = false,
		}
	end
	
	-- list of tracked node types (both professions)
	self.nodeTypes = {}
	
	-- resting state
	self.isResting = false
	
	-- living state
	self.alive = nil
	
	-- combat state
	self.inCombat = false
	
	-- flying state
	self.isFlying      = false
	self.isFlyingTimer = nil
	
	self.onTaxi        = false
	
	-- zone specific data
	self.zoneMap        = nil
	self.zoneID         = nil
	self.zoneFloor      = nil
	self.zoneName       = nil
	self.zoneLow        = nil
	self.zoneHigh       = nil
	
	self.pausedInZone        = false
	self.pausedForLevel      = false
	self.pausedWhenMapZoomed = false

	self.searchZones = {}
	
	self.isMicroDungeon = false
	
	-- cpu throttling
	self.totalElapsed    = 0.0
	self.updatePeriod    = 0.1
	self.updateTimes     = {}
	self.updateTotalTime = 0
	
	for i = 1, 100, 1 do
		self.updateTimes[i] = 0
	end
	
	-- timestamps of last visit at node location
	self.visited = {}
	
	self:RegisterChatCommand("magneto", "ChatCommand")
end

function Addon:OnEnable()
	self:SetupEvents()
	
	-- set up tracking spells
	self:UpdateTrackingSpells()
	
	-- set initial combat state
	self:UpdateRestingState()
	
	-- set initial combat state
	self:UpdateCombatState()
	
	-- set initial alive state
	self:UpdateAliveState()
	
	-- set initial tracking state
	self:UpdateTrackingState()
	
	-- set initial skill data
	self:UpdateSkill()
	self:UpdateLoggingSkill()
	
	-- initial update for label
	self:UpdateLabel()
	
	self:InitializeSettings()
end

function Addon:OnDisable()
	self:ShutdownEvents()

	MinimapButton:SetShow(false)
end

function Addon:OnOptionsReloaded()
	self:InitializeSettings()
end

function Addon:ChatCommand(input)
    if input then  
		args = NS:GetArgs(input)
		
		self:TriggerAction(unpack(args))
	        
        NS:ReleaseTable(args)
	else
		self:TriggerAction("help")
	end
end

function Addon:OnClick(button)
	if ( button == "RightButton" ) then 
		if IsShiftKeyDown() then
			-- unused
		elseif IsControlKeyDown() then
			-- unused
		elseif IsAltKeyDown() then
			-- unused
		else
			-- open options menu
			self:TriggerAction("menu")
		end
	elseif ( button == "LeftButton" ) then 
		if IsShiftKeyDown() then
			-- unused
		elseif IsControlKeyDown() then
			-- unused
		elseif IsAltKeyDown() then
			-- unused
		else
			-- toggle active
			self:TriggerAction("toggle")
		end
	end
end

function Addon:TriggerAction(action, ...)
    local args = NS:NewTable(...)

	action = type(action) == "string" and string.lower(action) or ""	

	if action == "menu" then
		-- open options menu
		InterfaceOptionsFrame_OpenToCategory(self.FULLNAME)
	elseif action == "debug" then
		if args[1] == "on" then
			self:Output("debug mode turned on")
			self.debug = true
		end
		if args[1] == "off" then
			self:Output("debug mode turned off")
			self.debug = false
		end
	elseif action == "on" then
		-- activate
		self:SetActive(true)
	elseif action == "off" then
		-- activate
		self:SetActive(false)
	elseif action == "force" then
		local magnitude, direction, sinDir, cosDir = self:GetForce()
		self:Output("Force magnitude: " .. tostring(magnitude))
		self:Output("Force direction: " .. tostring(direction) .. " (" ..tostring(cosDir) .. ", " .. tostring(sinDir) .. ")")
	elseif action == "zone" then
		local zone = self.zoneName
		self:Output("Zone: " .. tostring(zone))
		self:Output("Is Instance: " .. tostring(LibTourist:IsInstance(zone)))
		self:Output("Is Hostile: " ..  tostring(LibTourist:IsHostile(zone)) .. " - IsZoneFactionSet(hostile) " .. tostring(Options:IsZoneFactionSet("hostile")))
		self:Output("Is Contested " .. tostring(LibTourist:IsContested(zone)) .. " - IsZoneFactionSet(contested) " ..  tostring(Options:IsZoneFactionSet("contested")))
		self:Output("Is Sanctuary " .. tostring(LibTourist:IsSanctuary(zone)) .. " - IsZoneFactionSet(sanctuary) " .. tostring(Options:IsZoneFactionSet("sanctuary")))
		self:Output("Is City: " .. tostring(LibTourist:IsCity(zone)) .. " - IsZoneTypeSet(city) " .. tostring(Options:IsZoneTypeSet("city")))
		self:Output("Is PvP Zone: " .. tostring(LibTourist:IsPvPZone(zone)) .. " - IsZoneTypeSet(pvp) " .. tostring(Options:IsZoneTypeSet("pvp")))
	elseif action == "smooth" then
		self:Output("Update smooth.")
		self:UpdateSmoothForce()
		self:Output("Done.")
	elseif action == "paused" then
		self:Output("Is Paused: " .. tostring(self:IsPaused()))
		self:Output("Paused in Combat: " .. tostring(self:IsPausedInCombat()))
		self:Output("Paused while Resting: " ..  tostring(self:IsPausedWhileResting()))
		self:Output("Paused in Zone: " .. tostring(self.pausedInZone))
		self:Output("Paused for Level: " .. tostring(self.pausedForLevel))
		self:Output("Paused when Dead: " .. tostring(not self:IsAlive()))
		self:Output("Paused when Map zoomed: " .. tostring(self.pausedWhenMapZoomed))
	elseif action == "profession" then
		local profession = LibNTI:GetProfessionByLibName(args[1])
		
		if not profession or not self.PROFESSIONS[profession] then
			self:Output("Unknown profession '" .. tostring(args[1]) .. "'")
			
			return
		end
		
		self:Output("All " .. profession .. " nodes:")
		local currentExpansion = GetExpansionLevel()
		local profession = LibNTI.PROFESSION_MINING

		for expansion = 0, currentExpansion do
			local expansionName = _G["EXPANSION_NAME"..tostring(expansion)] or "unknown"

			self:Output(expansionName)
			
			local nodes = LibNTI:GetNodesForExpansion(profession, expansion)
			
			table.sort(nodes, NS.compareNodes)
			
			for index, nodeId in ipairs(nodes) do
				local nodeName  = LibNTI:GetName(profession, nodeId)
				
				self:Output(tostring(nodeName))
			end
		end
	elseif action == "mapinfo" then
		local mapName, textureWidth, textureHeight, isMicroDungeon, microDungeonName = GetMapInfo()
		self:Output("Mapinfo:" .. tostring(mapName) .. ", " .. tostring(textureWidth) .. ", " .. tostring(textureHeight) .. ", " .. tostring(isMicroDungeon) .. ", " .. tostring(microDungeonName) .. "; mzone .. " .. tostring(GetCurrentMapZone()))
	elseif action == "pos" then
		local x, y = GetPlayerMapPosition("player")
		self:Output("Player pos:" .. tostring(x) .. ", " .. tostring(y))
		local nx, ny = LibMapData:GetPlayerZonePosition()
		self:Output("Lib:Player npos:" .. tostring(nx) .. ", " .. tostring(ny))
	elseif action == "mapreset" then
		if _G.WorldMapFrame:IsVisible() then
			self:Output("SetMapToCurrentZone() aborted WorldMap is visible.")
		else
			SetMapToCurrentZone()
			self:Output("SetMapToCurrentZone() done.")
		end
	elseif action == "zoom" then
		if _G.WorldMapFrame:IsVisible() then
			self:Output("ZoomOut() aborted WorldMap is visible.")
		else
			ZoomOut()
			self:Output("ZoomOut() done.")
		end
	elseif action == "tracking" then
		for profession in pairs(Addon.PROFESSIONS) do
			if self.tracking[profession] then
				self:Output("Tracking info for " .. profession .. ": " .. tostring(self.tracking[profession].name) .. ", " .. tostring(self.tracking[profession].spellId) .. ", " .. tostring(self.tracking[profession].ready) .. ", " .. tostring(self.tracking[profession].active))
			else
				self:Output("No Tracking info for " .. profession)
			end
		end		
	elseif action == "skill" then
		self:UpdateSkill()
	elseif action == "logging" then
		self:UpdateLoggingSkill()
	elseif action == "toggle" then
		-- toggle operation
		self:ToggleActive()
	elseif action == "version" then
		-- print version information
		self:PrintVersionInfo()
	else -- if action == "help" then
		-- display help
		self:Output(L["Usage:"])
		self:Output(L["/magneto arg"])
		self:Output(L["Args:"])
		self:Output(L["menu - display options menu"])
		self:Output(L["on - activate operation"])
		self:Output(L["off - deactivate operation"])
		self:Output(L["toggle - toggle operation"])
		self:Output(L["version - display version information"])
		self:Output(L["help - display this help"])
	end
    
    NS:ReleaseTable(args)
end

function Addon:CreateTooltip(obj, autoHideDelay)
	Tooltip:Create(obj, autoHideDelay)
end

function Addon:RemoveTooltip()
	Tooltip:Remove()
end

function Addon:SetupEvents()
	self:RegisterMessage(ADDON .. "_SETTING_CHANGED", "UpdateSetting")	
	self:RegisterMessage("MAGNETO_INLET_ACTIVE_CHANGED", "ActiveInletUpdated")	
	self:RegisterMessage("MAGNETO_INLET_REGISTERED_CHANGED", "UpdateInletRegistered")	
	self:RegisterMessage("MAGNETO_OUTLET_REGISTERED_CHANGED", "UpdateOutletRegistered")	
	self:RegisterMessage("MAGNETO_ALGORITHM_REGISTERED_CHANGED", "UpdateAlgorithmRegistered")	

	-- repeating timer to keep track of flying state
	-- NOTE: could track mounted state and keep timer running only if actually mounted
	--       but one event every 0.5 sec doesn't seem to justify the effort
	self.isFlyingTimer = self:ScheduleRepeatingTimer("UpdateFlyingState", 0.5)

	-- register event to track zone changes
	LibMapData.RegisterCallback(Addon, "PlayerZoneChanged", "UpdateZone")
	
	-- register event to track spell changes (used to keep track of available tracking spells)
	--self:RegisterBucketEvent("SPELLS_CHANGED", 2, "UpdateTrackingSpells")
	self:RegisterEvent("SKILL_LINES_CHANGED", "UpdateTrackingSpells")

	-- register event to track skill changes
	self:RegisterEvent("CHAT_MSG_SKILL")
	
	-- register garrisons buildings events for keeping track of logging skill
    self:RegisterEvent("GARRISON_UPDATE", "UpdateLoggingSkill")
    self:RegisterEvent("GARRISON_BUILDING_ACTIVATED", "UpdateLoggingSkill")
    self:RegisterEvent("GARRISON_BUILDING_UPDATE", "UpdateLoggingSkill")
    self:RegisterEvent("GARRISON_BUILDING_REMOVED", "UpdateLoggingSkill")
	
	-- register events to track resting state
    self:RegisterEvent("PLAYER_UPDATE_RESTING", "UpdateRestingState")
	
	-- register events to track combat
	self:RegisterEvent("PLAYER_REGEN_ENABLED",  "UpdateCombatState", false)
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "UpdateCombatState", true)

	-- register events to track life state
	self:RegisterEvent("PLAYER_DEAD",    "UpdateAliveState")
	self:RegisterEvent("PLAYER_UNGHOST", "UpdateAliveState")
	self:RegisterEvent("PLAYER_ALIVE",   "UpdateAliveState")
	
	-- register event to track tracking states
	self:RegisterEvent("MINIMAP_UPDATE_TRACKING", "UpdateTrackingState")

	-- register event to track world map changes
	self:RegisterEvent("WORLD_MAP_UPDATE", "UpdateWorldMap")

	-- track level ups
	self:RegisterEvent("PLAYER_LEVEL_UP", "UpdatePlayerLevel")
	
	-- track enter world (loading, reloadui, porting etc.)
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "UpdatePlayerState")
end

function Addon:ShutdownEvents()
	self:UnregisterMessage(ADDON .. "_SETTING_CHANGED")	
	self:UnregisterMessage("MAGNETO_INLET_ACTIVE_CHANGED")	
	self:UnregisterMessage("MAGNETO_INLET_REGISTERED_CHANGED")	
	self:UnregisterMessage("MAGNETO_OUTLET_REGISTERED_CHANGED")	

	-- cancel repeating timer to keep track of flying state
	if self.isFlyingTimer then
		self:CancelTimer(self.isFlyingTimer)
		self.isFlyingTimer = nil
	end
	
	-- unregister event to track zone changes
	LibMapData.UnregisterCallback(Addon, "PlayerZoneChanged")

	-- unregister event to track spell changes
	--self:UnregisterEvent("SPELLS_CHANGED")
	--self:UnregisterAllBuckets()
	self:UnregisterEvent("SKILL_LINES_CHANGED")
	
	-- unregister event to track skill changes
	self:UnregisterEvent("CHAT_MSG_SKILL")

	-- register garrisons buildings events for keeping track of logging skill
    self:UnregisterEvent("GARRISON_UPDATE")
    self:UnregisterEvent("GARRISON_BUILDING_ACTIVATED")
    self:UnregisterEvent("GARRISON_BUILDING_UPDATE")
    self:UnregisterEvent("GARRISON_BUILDING_REMOVED")
	
	-- unregister events to track resting state
    self:RegisterEvent("PLAYER_UPDATE_RESTING", "UpdateRestingState")
	
	-- unregister events to track combat
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	
	-- unregister events to track life state
	self:UnregisterEvent("PLAYER_DEAD")
	self:UnregisterEvent("PLAYER_UNGHOST")
	self:UnregisterEvent("PLAYER_ALIVE")
	
	-- unregister event to track tracking states
	self:UnregisterEvent("MINIMAP_UPDATE_TRACKING")
	
	-- unregister event to track world map changes
	self:UnregisterEvent("WORLD_MAP_UPDATE")
	
	-- unregister level up tracking
	self:UnregisterEvent("PLAYER_LEVEL_UP")

	-- unregister enter world tracking (loading, reloadui, porting etc.)
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
end

function Addon:InitializeSettings()
	self:UpdateSmoothnessSettings()

	-- set up zone related settings
	self:UpdateZoneSettings()
	self:UpdateLevelSettings()

	-- initial node selection
	self:UpdateNodeTypeSelection()
	
	Inlets:SetActiveInlet(Options:GetSetting("inlet"))
	
	for name, outlet in Outlets:IterateOutlets() do
		Outlets:SetOutletActive(name, Options:IsOutletActive(name))
	end

	Algorithms:SetActiveAlgorithm(Options:GetSetting("algorithm"))
	
	MinimapButton:SetShow(Options:GetSetting("minimap"))
	
	self:UpdateIcon()
	self:UpdateState()	
end

-- calculation logic
function Addon:Process(elapsed)
	self.totalElapsed = self.totalElapsed + elapsed
	
	if self.totalElapsed > self.updatePeriod then
		self.totalElapsed = 0.0

		local pX, pY = LibMapData:GetPlayerZonePosition()

		if pX and pY and (pX ~= self.playerX or pY ~= self.playerY) then
			local startTime = GetTime()
			
			self:UpdateForce()
			
			local duration = GetTime() - startTime
			
			self.updateTotalTime = self.updateTotalTime - self.updateTimes[1] + duration
			
			tremove(self.updateTimes, 1)			
			tinsert(self.updateTimes, duration)						
			
			local maxCPULoad = Options:GetSetting("maxCPULoad")
			local minUpdatePeriod = Options:GetSetting("minUpdatePeriod")
			
			local avgTime = self.updateTotalTime / #self.updateTimes
			self.updatePeriod = avgTime / maxCPULoad
			
			if self.updatePeriod < minUpdatePeriod then
				self.updatePeriod = minUpdatePeriod
			end
		end		
	end
end

function Addon:UpdateForce()
	if self:GetState() ~= self.TRACKING then
		self:SetForce(0, PI, 0, 0, 0)
		return
	end

	local inlet = Inlets:GetActiveInlet()
	local algorithm = Algorithms:GetActiveAlgorithm()

	if not inlet or not algorithm then
		return
	end

	local pX, pY = GetPlayerMapPosition("player")

	self.playerX = pX
	self.playerY = pY

	local magnitude, direction, rating, totalCount, visitedCount = algorithm:CalculateForce(inlet, pX, pY, self.zoneName, self.zoneMap)
		
	self:SetForce(magnitude, direction, rating, totalCount, visitedCount)
end

--- events
function Addon:CHAT_MSG_SKILL(event, msg)
	if not msg then
		return
	end
	
	if msg:find(SKILL_UP_PATTERN) or msg:find(SKILL_LEARNED_PATTERN) then
		self:UpdateSkill()
	end
end

-- user functions
function Addon:PrintVersionInfo()
    self:Output(L["Version"] .. " " .. White(GetAddOnMetadata(ADDON, "Version")))
end

-- update functions
function Addon:Update()
	self:UpdateData()
	self:UpdateLabel()
end

function Addon:UpdateData()
	if self:GetState() == self.TRACKING then
		if not self.running then
			self.frame:SetScript("OnUpdate", function(self, elapsed) Addon:Process(elapsed) end)
			
			self.running = true
		end
	else
		if self.running then		
			self.frame:SetScript("OnUpdate", nil)
			
			self.playerX = nil
			self.playerY = nil
			
			self.running = false
			
			self:SendMessage("MAGNETO_FORCE_CHANGED", name)
		end
	end
end

function Addon:UpdateLabel()
	local label = ""
	local state = self:GetState()
	
	if state == self.INACTIVE then
		label = label .. NS:Colorize("GrayOut", L["D"])
	elseif state == self.IDLE then
		label = label .. NS:Colorize("White", L["I"])
	elseif state == self.PAUSED then
		label = label .. NS:Colorize("Yellow", L["P"])
	elseif state == self.TRACKING then
		label = label .. NS:Colorize("Green", L["T"])
	end

	label = label .. ": "
	
	for profession, tracking in pairs(self.tracking) do
		local abbrev = self.PROFESSIONS[profession].abbrev
		
		if tracking.ready then
			if tracking.active then
				label = label .. NS:Colorize("Green", abbrev)
			else
				label = label .. NS:Colorize("Red", abbrev)
			end
		else
			label = label .. NS:Colorize("GrayOut", abbrev)
		end
	end
	
	DataBroker:SetText(label)
end

function Addon:UpdateIcon()
	local icon = self:GetIcon()
	
	DataBroker:SetIcon(icon)
	MinimapButton:SetIcon(icon)
end

function Addon:UpdateState()
	local state = self.INACTIVE
	
	if self:IsActive() and Inlets:IsInletOperational() then
		if self:IsAnyTrackingActive() then
			if self:IsPaused() then
				state = self.PAUSED
			else
				state = self.TRACKING
			end
		else
			state = self.IDLE
		end
	end
	
	if state ~= self.state then
		self.state = state

		self:Update()
		self:UpdateForce()
	end
end

function Addon:UpdatePlayerLevel()
	self.playerLvl = UnitLevel("player")
	
	self:UpdateLevelSettings()
end

function Addon:UpdateZone(event, mapID, floor, map, isMicroDungeon)
	self.zoneMap   = map
	self.zoneFloor = floor
	self.zoneID    = mapID
	self.zoneName  = GetMapNameByID(self.zoneID)
	
	self.zoneLow, self.zoneHigh = LibTourist:GetLevel(self.zoneName)

	-- make sure force is refreshed if zone change happens because of world map close when standing still
	self.playerX = nil
	self.playerY = nil
	
	self:SetIsInMicroDungeon(isMicroDungeon)
	
	self:UpdateZoneSettings()
	self:UpdateLevelSettings()

	self:UpdateSearchZones()
	
	self:UpdateNodeTypeSelection()
end

function Addon:UpdateWorldMap(event)
	local zoomed = false

	local _, _, _, isMicroDungeon = GetMapInfo()
	
	if isMicroDungeon then
		if _G.WorldMapFrame:IsVisible() then
			zoomed = true
		end
	else
		zoomed = GetCurrentMapZone() == 0
	end
	
	if zoomed ~= self.pausedWhenMapZoomed then
		self.pausedWhenMapZoomed = zoomed
		
		self:UpdateState()	
	end
end

function Addon:UpdateLevelSettings()
	local pause = false

	local level = Options:GetZoneLevel()
	
	if level == self.LEVEL_RANGE then
		-- if zoneHigh == 0 the zone is free of restrictions
		if self.zoneHigh > 0 and (self.playerLvl > self.zoneHigh or self.playerLvl < self.zoneLow) then
			pause = true
		end
	elseif level == self.LEVEL_LOW then
		if self.playerLvl < self.zoneLow then
			pause = true
		end
	end
	
	if pause ~= self.pausedForLevel then
		self.pausedForLevel = pause
		
		self:UpdateState()
	end
end

function Addon:UpdateZoneSettings()
	local pause = self:IsZonePaused(self.zoneName)
	
	if pause ~= self.pausedInZone then
		self.pausedInZone = pause
		
		self:UpdateState()
	end
end

function Addon:UpdateNodeTypeSelection()
	NS:ClearTable(self.nodeTypes)

	local zoneSpecific = Options:IsZoneSpecific()
	
	for profession in pairs(self.PROFESSIONS) do
		if self.tracking[profession].active then
			for nodeType in LibNTI:IterateNodes(profession) do
				if not Options:IsNodeTypeBlacklisted(nodeType) then
					local expansion = LibNTI:GetExpansion(profession, nodeType)
					
					if Options:IsExpansionSet(expansion) then
						local zoneOk = LibNTI:SpawnsInZone(profession, nodeType, self.zoneID)
						
						-- handle multi zone if nodes are not restricted to current zone
						if not zoneOk and not zoneSpecific then
							for zoneId, zone in self:IterateSearchZones() do
								if zone ~= self.zoneName and LibNTI:SpawnsInZone(profession, nodeType, zoneId) then
									zoneOk = true
									
									break
								end
							end
						end
						
						if zoneOk then
							self.nodeTypes[nodeType] = true
						end
					end
				end 
			end
		end
	end
	
	self:SendMessage("MAGNETO_NODE_TYPES_CHANGED")
end

function Addon:UpdatePlayerState()
	self:UpdateCombatState()
	self:UpdateAliveState()
	self:UpdateRestingState()
end

function Addon:UpdateCombatState(inCombat)
	if inCombat == nil then
		inCombat = UnitAffectingCombat("player")
	end
	
	if inCombat ~= self.inCombat then
		self.inCombat = inCombat
	
		self:UpdateState()	
	end
end

function Addon:UpdateAliveState()
	local alive = not UnitIsDeadOrGhost("player")
	
	if alive ~= self.alive then
		self.alive = alive
	
		self:UpdateState()	
	end
end

function Addon:UpdateRestingState()
    local isResting = IsResting() or false	
		
	if isResting ~= self.isResting then
		self.isResting = isResting
	
		self:UpdateState()	
	end
end

function Addon:UpdateFlyingState()
	local onTaxi = UnitOnTaxi("player") and true or false

	if flying ~= self.onTaxi then
		self.onTaxi = onTaxi
	
		self:UpdateState()	
	end
	
	local flying = IsFlying() and true or false
	
	if flying ~= self.isFlying then
		self.isFlying = flying
	
		self:UpdateSearchZones()	
	end
end

function Addon:UpdateTrackingSpells()
	for profession in pairs(self.PROFESSIONS) do
		local trackingSpell = LibNTI:GetTrackingSpell(profession)
		
		self.tracking[profession].spellId = trackingSpell
		self.tracking[profession].name    = GetSpellInfo(trackingSpell)
	end

	-- NOTE: why does IsSpellKnown(spellID) does not work for Find Minerals/Herbs (2580/2383)?
	local changed   = false
	
	local knownTracking = {}

	for i = 1, GetNumTrackingTypes(), 1 do
		local name, _, active, _ = GetTrackingInfo(i)
		
		knownTracking[name] = true
	end	

	for profession, trackingInfo in pairs(self.tracking) do
		local ready = knownTracking[trackingInfo.name] and true or false
		
		if trackingInfo.ready ~= ready then
			trackingInfo.ready = ready
			
			changed = true
		end
	end
		
	if changed then
		self:UpdateTrackingState()	
		
		self:UpdateState()
		self:UpdateLabel()
		Tooltip:Refresh()		
	end
end

function Addon:UpdateTrackingState()
	local changed   = false

	for i = 1, GetNumTrackingTypes(), 1 do
		local name, _, active, _ = GetTrackingInfo(i)
		
		active = active and true or false
		
		for profession in pairs(Addon.PROFESSIONS) do
			if name == self.tracking[profession].name then
				if self.tracking[profession].active ~= active then
					self.tracking[profession].active = active
					changed = true
				end
			end
		end		
	end

	if changed then
		self:UpdateNodeTypeSelection()
		
		self:UpdateState()
		self:UpdateLabel()
		Tooltip:Refresh()
	end
end

function Addon:UpdateSkill()
	local prof1, prof2, archaeology, fishing, cooking, firstAid = GetProfessions()

	self:UpdateSkillOfProfession(prof1)	
	self:UpdateSkillOfProfession(prof2)
	self:UpdateSkillOfProfession(fishing)
end

function Addon:UpdateSkillOfProfession(profId)
	if not profId then
		return
	end
	
	local name, _, rank, maxRank, _, _, skillLine = GetProfessionInfo(profId)
		
	self:SetSkill(self.mapSkillId[skillLine], rank)
end

function Addon:UpdateLoggingSkill()
	local rank = Options:GetSetting("loggingskill")

	if self.tracking[LibNTI.PROFESSION_LOGGING].ready then
		local plots = GetPlots(LE_FOLLOWER_TYPE_GARRISON_6_0)
		
		if plots and next(plots) then
			for key, value in pairs(plots) do
				local buildingID = GetOwnedBuildingInfo(value.id)

				if lumbermillSize[buildingID] then
					rank = lumbermillSize[buildingID]
				
					break
				end		
			end
		end		
	else
		rank = nil
	end
	
	Options:SetSetting("loggingskill", rank)
	
	rank = rank or "--"
	
	self:SetSkill(LibNTI.PROFESSION_LOGGING, rank)
end

-- TODO: check if lib tourist is updated for bordering zones via flight mount
function Addon:UpdateSearchZones()
	NS:ClearTable(self.searchZones)

	if not LibTourist:IsInstance(self.zoneName) then
		self.searchZones[self.zoneID] = self.zoneName
	
		local multiType = Options:GetMultiZone()
		
		local isMulti = multiType == self.MULTI_ALWAYS or (multiType == self.MULTI_FLYING and self.isFlying)
		
		if isMulti then
			for zone in LibTourist:IterateBorderZones(self.zoneName, true) do
				if not self:IsZonePaused(zone) and not LibTourist:IsInstance(zone) then
					local zoneMap = LibTourist:GetTexture(zone)
					local zoneId  = LibMapData:GetMapIDFromFile(zoneMap)
					
					self.searchZones[zoneId] = zone
				end
			end
		end
	end
end

function Addon:UpdateSmoothnessSettings()
	local duration = Options:GetSetting("smoothness")
	
	self.smoothStep = PI / (1 + duration*20)
	
	self.smoothMagnitude = self.forceMagnitude
	self.smoothDirection = self.forceDirection
	self.smoothSinDir    = self.forceSinDir
	self.smoothCosDir    = self.forceCosDir
end

function Addon:UpdateInletRegistered(event, name, inlet, registered)
	if registered then
		local active = self:GetSetting("inlet")
		
		if name == active then
			Inlets:SetActiveInlet(name)
		end
	end
	
	Options:Update()
end

function Addon:UpdateOutletRegistered(event, name, outlet, registered)
	if registered then
		Options:RegisterOutletOptions(outlet)
	
		Outlets:SetOutletActive(name, Options:IsOutletActive(name))
	else
		Options:UnregisterOutletOptions(outlet)
	end
	
	Options:Update()
end

function Addon:UpdateAlgorithmRegistered(event, name, algorithm, registered)
	if registered then
		Options:RegisterAlgorithmOptions(algorithm)
		
		local active = self:GetSetting("algorithm")
		
		if name == active then
			Algorithms:SetActiveAlgorithm(name)
		end		
	else
		Options:UnregisterAlgorithmOptions(algorithm)
	end
	
	Options:Update()
end

-- settings
function Addon:UpdateSetting(event, setting, value, old, category, ...)
	local handler = updateHandler[setting]

	if handler == "UpdateActiveSetting" then
		self:UpdateState()
		self:UpdateIcon()
		Tooltip:Refresh()
	elseif handler == "UpdateMinimapSetting" then
		MinimapButton:SetShow(value)
	elseif handler == "UpdateInletSetting" then
		Inlets:SetActiveInlet(value)
	elseif handler == "UpdateOutletSetting" then
		local name = category
		
		Outlets:SetOutletActive(name, value)
	elseif handler == "UpdateAlgorithmSetting" then
		Algorithms:SetActiveAlgorithm(value)
	elseif handler == "UpdateSmoothnessSetting" then
		self:UpdateSmoothnessSettings()
	elseif handler == "UpdateStateSetting" then
		self:UpdateState()
	elseif handler == "UpdateZoneLevelSetting" then
		self:UpdateLevelSettings()
	elseif handler == "UpdateMultiZoneSetting" then
		self:UpdateSearchZones()		
		self:UpdateNodeTypeSelection()
	elseif handler == "UpdateZoneSetting" then
		self:UpdateZoneSettings()
	elseif handler == "UpdateNodeSetting" then
		self:UpdateNodeTypeSelection()
	end
end

function Addon:GetSetting(setting)
	return Options:GetSetting(setting)
end

function Addon:ToggleActive()
	self:SetActive(not self:IsActive())
end

function Addon:SetActive(active)
	active = active and true or false
	
	Options:SetSetting("active", active)
end

function Addon:ActiveInletUpdated(name)
	self:UpdateState()
end

-- utilities
function Addon:ToggleTracking(profession)
	local trackingName = self.tracking[profession] and self.tracking[profession].name or nil

	if not trackingName then
		return
	end
	
	local enable = not self.tracking[profession].active
	
	for i = 1, GetNumTrackingTypes(), 1 do
		local name = GetTrackingInfo(i)

		if name == trackingName then
			SetTracking(i, enable)
			
			return
		end
	end
end

function Addon:SetForce(magnitude, direction, rating, nodeCount, visitedCount)
	magnitude    = magnitude or 0
	direction    = direction or 0
	rating       = rating or 0
	nodeCount    = nodeCount or 0
	visitedCount = visitedCount or 0

	while direction > TWOPI do
		direction = direction - TWOPI
	end
	
	while direction < 0 do
		direction = direction + TWOPI
	end
	
	if magnitude ~= self.forceMagnitude or direction ~= self.forceDirection then
		self.forceMagnitude = magnitude
		self.forceDirection = direction
		
		self.forceSinDir    = sin(direction)
		self.forceCosDir    = cos(direction)
		
		if direction == 0 then
			rating = 0
		elseif rating < 1 then
			rating = 1
		elseif rating > 5 then
			rating = 5
		else
			rating = floor(rating)
		end		
		
		self.forceRating       = rating
		self.forceTotalCount   = nodeCount
		self.forceVisitedCount = visitedCount
		
		-- if smoothness create sequence of updates with resolution of 0.1s
		if Options:GetSetting("smoothness") > 0 then 
			self:UpdateSmoothForce()
		else
			self:SendMessage("MAGNETO_FORCE_CHANGED", self.forceMagnitude, self.forceDirection, self.forceSinDir, self.forceCosDir)
		end
	end
end

function Addon:UpdateSmoothForce()
	-- switch on/off without smooth update
	if self.forceMagnitude == 0 or self.smoothMagnitude == 0 then
		self.smoothMagnitude = self.forceMagnitude
		self.smoothDirection = self.forceDirection
		self.smoothSinDir = self.forceSinDir
		self.smoothCosDir = self.forceCosDir
		
		self:SendMessage("MAGNETO_FORCE_CHANGED", self.smoothMagnitude, self.smoothDirection, self.smoothSinDir, self.smoothCosDir)
		
		return
	end
	
	local distance = self.forceDirection - self.smoothDirection

	local step    = 0
	local stretch = 1
	
	if distance > 0 then
		if distance < self.smoothStep then
			self.smoothDirection = self.forceDirection
			self.smoothMagnitude = self.forceMagnitude
		elseif distance > PI then
			step = -self.smoothStep
			
			if self.smoothDirection < self.smoothStep then
				step = step + TWOPI
			end
			
			stretch = self.smoothStep / (TWOPI-distance)
		else
			step    = self.smoothStep
			stretch = self.smoothStep / distance
		end		
	elseif distance < 0 then
		if -distance < self.smoothStep then
			self.smoothDirection = self.forceDirection
			self.smoothMagnitude = self.forceMagnitude
		elseif distance < -PI then
			step = self.smoothStep

			if TWOPI - self.smoothDirection < self.smoothStep then
				step = step - TWOPI
			end

			stretch = self.smoothStep / (distance + TWOPI)
		else
			step    = -self.smoothStep
			stretch = self.smoothStep / -distance
		end
	end

	if step ~= 0 then
		self.smoothDirection = self.smoothDirection + step

		local difference = self.forceMagnitude - self.smoothMagnitude
		self.smoothMagnitude = self.smoothMagnitude + difference * stretch
	end
	
	if distance == 0 then
		-- if only magnitude changes: just up date
		if self.smoothMagnitude ~= self.forceMagnitude then
			self.smoothMagnitude = self.forceMagnitude
		end
	else
		self.smoothSinDir = sin(self.smoothDirection)
		self.smoothCosDir = cos(self.smoothDirection)
	end
	
	self:SendMessage("MAGNETO_FORCE_CHANGED", self.smoothMagnitude, self.smoothDirection, self.smoothSinDir, self.smoothCosDir)
	
	if self.smoothDirection ~= self.forceDirection then
		self:ScheduleTimer("UpdateSmoothForce", 0.05)
	end	
end

function Addon:SetVisited(zone, encLoc, timestamp)
	if not self.visited[zone] then
		self.visited[zone] = {}
	end
	
	self.visited[zone][encLoc] = timestamp
	
	if not self.visitedTimer then
		self.visitedTimer = self:ScheduleTimer("CleanupVisits", 60)
	end
end

function Addon:CleanupVisits(force)
	if self.visitedTimer then
		self:CancelTimer(self.visitedTimer)
		self.visitedTimer = nil
	end
	
	local oldestValid = time() - self.MAX_NODE_AGE
	
	for zone, data in pairs(self.visited) do
		for encLoc, timestamp in pairs(data) do
			if timestamp < oldestValid or force then
				self.visited[zone][encLoc] = nil			
			end
		end
		
		if not next(self.visited[zone]) then
			self.visited[zone] = nil
		end
	end

	-- if there is still data in the table reschedule timer
	if next(self.visited) then
		self.visitedTimer = self:ScheduleTimer("CleanupVisits", 60)
	end
end

function Addon:IsZonePaused(zone)
	if LibTourist:IsInstance(zone) then
		return true
	elseif LibTourist:IsHostile(zone) and not Options:IsZoneFactionSet("hostile") then
		return true
	elseif LibTourist:IsContested(zone) and not Options:IsZoneFactionSet("contested") then
		return true
	elseif LibTourist:IsSanctuary(zone) and not Options:IsZoneFactionSet("sanctuary") then
		return true
	elseif LibTourist:IsCity(zone) and not Options:IsZoneTypeSet("city") then
		return true
	elseif LibTourist:IsPvPZone(zone) and not Options:IsZoneTypeSet("pvp") then
		return true
	end
	
	return false
end

function Addon:SetSkill(profession, rank)
	if not profession or not Addon.PROFESSIONS[profession] then
		return
	end
	
	if self.skill[profession] ~= rank then
		self.skill[profession] = rank
		
		self:UpdateNodeTypeSelection()
	end
end

function Addon:SetIsInMicroDungeon(inMicroDungeon)
	inMicroDungeon = inMicroDungeon and true or false
	
	if self.isMicroDungeon ~= inMicroDungeon then
		self.isMicroDungeon = inMicroDungeon
		
		self:UpdateState()
	end
end

function Addon:Output(msg)
	if msg ~= nil and DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage(self.MODNAME..": "..msg, 0.6, 1.0, 1.0)
	end
end

-- test
function Addon:Debug(msg)
	if self.debug and msg ~= nil and DEFAULT_CHAT_FRAME then
		DEFAULT_CHAT_FRAME:AddMessage(self.MODNAME .. " (dbg): " .. msg, 1.0, 0.37, 0.37)
	end
end

-- get addon data
function Addon:GetIcon()
	if self:IsActive() then
		return ICON
	else
		return ICON_DISABLED
	end
end

function Addon:GetState()
	return self.state
end

function Addon:IterateSearchZones()
	return pairs(self.searchZones)
end

function Addon:IterateNodeTypeSelection()
	return pairs(self.nodeTypes)
end

function Addon:GetSkill(profession)
	-- since garrison API apparently doesn't provide correct info right after PLAYER_ENTERING_WORLD
	-- we have to recheck logging if tracking exists but no skill level is set
	if profession == LibNTI.PROFESSION_LOGGING and self.tracking[profession].ready and self.skill[profession] == "--" then
		self:UpdateLoggingSkill()
	end

	return self.skill[profession]
end

-- returns magnitude, direction (rad), sin(dir), cos(dir), rating
function Addon:GetForce()
	if not self.running then
		return 0, 0, 1, 0, 0, 0, 0
	elseif Options:GetSetting("smoothness") > 0 then
		return self.smoothMagnitude, self.smoothDirection, self.smoothSinDir, self.smoothCosDir, self.forceRating, self.forceTotalCount, self.forceVisitedCount
	else
		return self.forceMagnitude, self.forceDirection, self.forceSinDir, self.forceCosDir, self.forceRating, self.forceTotalCount, self.forceVisitedCount
	end
end

function Addon:GetVisited(zone, encLoc)
	return self.visited[zone] and self.visited[zone][encLoc]
end

function Addon:GetAge(zone, encLoc, now, maxAge)
	local timestamp = self:GetVisited(zone, encLoc)
	local age = self.MAX_NODE_AGE
	
	if now and timestamp then
		age = now - timestamp
	end
	
	return age <= maxAge and age or maxAge
end

function Addon:GetZoneInfo()
	return self.zoneName, self.zoneID, self.zoneMap, self.zoneFloor, self.zoneLow, self.zoneHigh
end

function Addon:IsPaused()
	return self:IsPausedInCombat() or self:IsPausedWhileResting() or self.pausedInZone or self.pausedForLevel or self.pausedWhenMapZoomed or self:IsOnTaxi() or not self:IsAlive() or self:IsInMicroDungeon()
end

function Addon:IsPausedInCombat()
	return Options:GetSetting("pauseInCombat") and self.inCombat
end

function Addon:IsAlive()
	return self.alive
end

function Addon:IsInMicroDungeon()
	return self.isMicroDungeon
end

function Addon:IsOnTaxi()
	return self.onTaxi
end

function Addon:IsPausedWhileResting()
	return Options:GetSetting("pauseWhenResting") and self.isResting
end

function Addon:IsActive()
	return Options:GetSetting("active")
end

function Addon:IsOperational()
	return self:IsActive() and Inlets:IsInletOperational()
end

function Addon:IsTracking(profession)
	return self.tracking[profession] and self.tracking[profession].active or false
end

function Addon:IsAnyTrackingActive()
	for profession in pairs(Addon.PROFESSIONS) do
		if self:IsTracking(profession) then
			return true
		end
	end
	
	return false
end

function Addon:IterateTrackingSkills()
	return pairs(self.tracking)
end
