-- TODO: 
-- decouple setting name from storage name and use flat interface
-- enrich with sanity checks and path info (for event)
-- maybe make pathinfo accessible via function and skip the event args
-- how to handle variable lists? like blacklist or outlets? (setting, value, old, specific -> e.g. 'blacklist', false, true, 1024)
-- SetSetting(setting, value, specific)
-- GetSetting(setting, specific) 
-- iterate options by category or all (subcategories?)
-- prob: performance will suffer for frequent access -> addon needs to copy the data it requires

local _G = _G

-- addon name and namespace
local ADDON, NS = ...

local Addon = LibStub("AceAddon-3.0"):GetAddon(ADDON)

-- the Options module
local Options = Addon:NewModule("Options", "AceEvent-3.0")

-- local functions
local ipairs   = ipairs 
local pairs    = pairs 

local GetExpansionLevel = _G.GetExpansionLevel

local _

local LibStub   = LibStub

-- config libraries
local AceConfig 		= LibStub:GetLibrary("AceConfig-3.0")
local AceConfigReg 		= LibStub:GetLibrary("AceConfigRegistry-3.0")
local AceConfigDialog	= LibStub:GetLibrary("AceConfigDialog-3.0")

-- translations
local L = LibStub:GetLibrary("AceLocale-3.0"):GetLocale(ADDON)

-- LibNodeTypeInfo
local LibNTI = LibStub:GetLibrary("LibNodeTypeInfo-1.0")

-- modules
local Inlets        = nil
local Outlets       = nil
local Algorithms    = nil

-- local variables
local defaults = {
	profile = {
		active              = true,
		visitingDistance    = 80,
		minNodeAge          = 600,
		maxNodeAge          = 3600,
		smoothness          = 1,
		pauseWhenResting    = true,
		pauseInCombat       = true,
		maxCPULoad          = 0.05,
		minUpdatePeriod     = 0.1,
		nodes               = {
			zoneSpecific = false,
			expansions = {},
			blacklist = {},
		},
		zones = {
			level = Addon.LEVEL_ALL,
			multi = Addon.MULTI_NEVER,
			factions = {
				hostile   = true,
				contested = true,
				sanctuary = true,
			},
			types = {
				city      = false,
				pvp       = true,
			},
		},
		tooltip = {},
		hideHint  = false,
        minimap   = false,
		inlet     = "none",
		outlets   = {},
		algorithm = "Arc",
		algorithms = {},
	}
}

-- outlet order
local outletOrder = 0

local inletOptions = {}

local algorithmOptions = {}

--local constants
local MIN_VISITING_DISTANCE =  20
local MAX_VISITING_DISTANCE = 180

local PI = math.pi

local levelRangeOptions = {
	[Addon.LEVEL_ALL]   = L[Addon.LEVEL_ALL], 
	[Addon.LEVEL_LOW]   = L[Addon.LEVEL_LOW], 
	[Addon.LEVEL_RANGE] = L[Addon.LEVEL_RANGE], 
}

local multiZoneOptions = {
	[Addon.MULTI_NEVER]  = L[Addon.MULTI_NEVER], 
	[Addon.MULTI_ALWAYS] = L[Addon.MULTI_ALWAYS], 
	[Addon.MULTI_FLYING] = L[Addon.MULTI_FLYING], 
}

local professionConfig = {
	[LibNTI.PROFESSION_MINING] = {
		key = "mining",
		name = L["Mining"],
		desc = L["Select minerals to track."],
	},
	[LibNTI.PROFESSION_HERBALISM] = {
		key = "herbs",
		name = L["Herbalism"],
		desc = L["Select herbs to track."],
	},
	[LibNTI.PROFESSION_FISHING] = {
		key = "fish",
		name = L["Fishing"],
		desc = L["Select fish schools to track."],
	},
	[LibNTI.PROFESSION_LOGGING] = {
		key = "logging",
		name = L["Logging"],
		desc = L["Select lumber to track."],
	},
} 

-- module handling
function Options:OnInitialize()
	-- set module references
	Inlets     = Addon:GetModule("Inlets")
	Outlets    = Addon:GetModule("Outlets")
	Algorithms = Addon:GetModule("Algorithms")

	-- options
	self.options = {}
	
	self.db = LibStub:GetLibrary("AceDB-3.0"):New(Addon.MODNAME.."_DB", defaults, "Default")
	
	local currentExpansion = GetExpansionLevel()

	for expansion = 0, currentExpansion do
		if self.db.profile.nodes.expansions[expansion] == nil then
			self.db.profile.nodes.expansions[expansion] = true
		end
	end	
	
	self:Setup()
	
	-- profile support
	self.options.args.profile = LibStub:GetLibrary("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied",  "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset",   "OnProfileChanged")

	AceConfigReg:RegisterOptionsTable(Addon.FULLNAME, self.options)
	AceConfigDialog:AddToBlizOptions(Addon.FULLNAME)	
end

function Options:OnEnable()
	-- empty
end

function Options:OnDisable()
	-- empty
end

function Options:OnProfileChanged(event, database, newProfileKey)
	self.db.profile = database.profile
	
	Addon:OnOptionsReloaded()
end

-- setup
function Options:Setup()	
    self.options = {
		handler = self,
		type = 'group',
		args = {
			general = {
				type = 'group',
				name = L["General"],
				order = 10,
				desc = L["Setup general options."],
				args = {
					inlet = { 
						type = 'select',
						name = L["Data Source"],
						desc = L["Select data source for addon."],
						order = 1,
						get = function()
							local inlet = self:GetSetting("inlet")
							
							inlet = Inlets:HasInlet(inlet) and inlet or "none"
							
							return inlet
						end,	
						set = function(info, inlet)
							self:SetSetting("inlet", inlet)
						end,
						values = "GetInletOptions",
					},
					visitingDistance = {
						type = 'range',
						name = L["Visiting Distance"],
						desc = L["Within this distance nodes are considered visited."],
						order = 2,
						get = function() 
							return self:GetSetting("visitingDistance")
						end,
						set = function(info, value) 
							self:SetSetting("visitingDistance", value)
						end,
						min = MIN_VISITING_DISTANCE,
						max = MAX_VISITING_DISTANCE,
						step = 1,
					},
					minAge = {
						type = 'range',
						name = L["Min. Visiting Age"],
						desc = L["Before this duration since last visit the node is ignored in the calculation."],
						order = 3,
						get = function() 
							return self:GetSetting("minNodeAge")/60
						end,
						set = function(info, value) 
							self:SetSetting("minNodeAge", value*60)
						end,
						min = Addon.MIN_NODE_IGNORE_TIME/60,
						max = Addon.MAX_NODE_IGNORE_TIME/60,
						step = 1,
					},
					maxAge = {
						type = 'range',
						name = L["Max. Visiting Age"],
						desc = L["After this duration since last visit the node is considered fresh again."],
						order = 3,
						get = function() 
							return self:GetSetting("maxNodeAge")/60
						end,
						set = function(info, value) 
							self:SetSetting("maxNodeAge", value*60)
						end,
						min = Addon.MIN_NODE_AGE/60,
						max = Addon.MAX_NODE_AGE/60,
						step = 1,
					},
					smoothness = {
						type = 'range',
						name = L["Smoothness"],
						desc = L["Smoothing out jerky direction changes. Time it takes for the force to invert direction."],
						order = 3,
						get = function() 
							return self:GetSetting("smoothness")
						end,
						set = function(info, value) 
							self:SetSetting("smoothness", value)
						end,
						min = 0,
						max = 2,
						step = 0.05,
					},
					whenResting = {
						type = 'toggle',
						name = L["Pause When Resting"],
						desc = L["Pause operation while resting."],
						order = 5,
						get  = function() return self:GetSetting("pauseWhenResting") end,
						set  = function()
							self:ToggleSetting("pauseWhenResting") 
						end,
					},				
					inCombat = {
						type = 'toggle',
						name = L["Pause In Combat"],
						desc = L["Pause operation while in combat."],
						order = 6,
						get  = function() return self:GetSetting("pauseInCombat") end,
						set  = function()
							self:ToggleSetting("pauseInCombat") 
						end,
					},				
				},
			},
			algorithms = {
				type = 'group',
				name = L["Calculation"],
				desc = L["Manage modules for force calculation."],
				order = 20,
				args = {
					algorithm = { 
						type = 'select',
						name = L["Algorithm"],
						desc = L["Select algorithm for calculation of Magneto force and direction."],
						order = 1,
						get = function()
							local algorithm = self:GetSetting("algorithm")
							
							algorithm = Algorithms:HasAlgorithm(algorithm) and algorithm or "none"
							
							return algorithm
						end,	
						set = function(info, algorithm)
							self:SetSetting("algorithm", algorithm)
						end,
						values = "GetAlgorithmOptions",
					},
					-- algorithm specific settings filled dynamically
				},
			},
			outlets = {
				type = 'group',
				name = L["Display"],
				desc = L["Manage display outlets."],
				order = 30,
				args = {
					-- filled dynamically
				},
			},		
			nodes = {
				type = 'group',
				name = L["Nodes"],
				order = 40,
				desc = L["Select nodes to include."],
				args = {
					zone = {
						type = 'toggle',
						name = L["Zone Specific"],
						desc = L["Include node types from the current zone only. When using multi-zone option only node types that are present in the current zone will be considered in neighboring zones."],
						order = 1,
						get  = function() 
							return self:IsZoneSpecific()
						end,
						set  = function()
							self:ToggleZoneSpecific()
						end,
					},	
					expansions = {
						type = "description",
						name = NS:Colorize("Yellow", L["Expansions"]) .. ": " .. L["Include node types that belong to the the specified expansion."],
						order = 2,
					},
					-- expansions for professions filled in self:SetupExpansionOptions()
					-- blacklists for professions filled in self:SetupBlacklistOptions()
				},
			},
			zones = {
				type = 'group',
				name = L["Zones"],
				order = 50,
				desc = L["Select zones for operation."],
				args = {
					level = { 
						type = 'select',
						name = L["Level"],
						desc = L["Include zones based on level range."],
						order = 1,
						get = function() 
							return self:GetZoneLevel()
						end,	
						set = function(info, key) 
							self:SetZoneLevel(key)
						end,
						values = levelRangeOptions,
					},
					multi = { 
						type = 'select',
						name = L["Multi-Zone"],
						desc = L["Include data from neighbouring zones."],
						order = 2,
						get = function() 
							return self:GetMultiZone()
						end,	
						set = function(info, key) 
							self:SetMultiZone(key)
						end,
						values = multiZoneOptions,
					},
					factions = {
						type = "description",
						name = NS:Colorize("Yellow", L["Factions"]) .. ": " .. L["Tracking while in zone of of selected factions."],
						order = 3,
					},
					hostile = {
						type = 'toggle',
						name = L["Hostile"],
						desc = L["Tracking while in zones of opposite faction."],
						order = 4,
						get  = function() 
							return self:IsZoneFactionSet("hostile")
						end,
						set  = function()
							self:ToggleZoneFaction("hostile")
						end,
					},				
					contested = {
						type = 'toggle',
						name = L["Contested"],
						desc = L["Tracking while in contested zone."],
						order = 6,
						get  = function() 
							return self:IsZoneFactionSet("contested")
						end,
						set  = function()
							self:ToggleZoneFaction("contested")
						end,
					},				
					sanctuary = {
						type = 'toggle',
						name = L["Sanctuary"],
						desc = L["Tracking while in sanctuaries."],
						order = 7,
						get  = function() 
							return self:IsZoneFactionSet("sanctuary")
						end,
						set  = function()
							self:ToggleZoneFaction("sanctuary")
						end,
					},									
					types = {
						type = "description",
						name = NS:Colorize("Yellow", L["Types"]) .. ": " .. L["Tracking while in zone of selected types. The addon is always paused in instances, arenas and battlegrounds."],
						order = 8,
					},
					city = {
						type = 'toggle',
						name = L["City"],
						desc = L["Tracking while in major cities."],
						order = 9,
						get  = function() 
							return self:IsZoneTypeSet("city")
						end,
						set  = function()
							self:ToggleZoneType("city")
						end,
					},	
					pvp = {
						type = 'toggle',
						name = L["PVP"],
						desc = L["Tracking while in pvp zones."],
						order = 11,
						get  = function() 
							return self:IsZoneTypeSet("pvp")
						end,
						set  = function()
							self:ToggleZoneType("pvp")
						end,
					},														
				},
			},
			cpu = {
				type = 'group',
				name = L["CPU"],
				desc = L["Manage settings to reduce cpu load."],
				order = 60,
				args = {
					cpuload = {
						type = 'range',
						name = L["Max. CPU Load"],
						desc = L["Maximum cpu load the addon is allowed to produce."],
						order = 1,
						get = function() 
							return self:GetSetting("maxCPULoad") * 100 
						end,
						set = function(info, value) 
							self:SetSetting("maxCPULoad", value/100)
						end,
						min = 1,
						max = 100,
						step = 1,
					},
					minperiod = {
						type = 'range',
						name = L["Min. Update Period"],
						desc = L["Lower time limit for data updates."],
						order = 2,
						get = function() 
							return self:GetSetting("minUpdatePeriod")
						end,
						set = function(info, value) 
							self:GetSetting("minUpdatePeriod", value)
						end,
						min = 0.05,
						max = 1,
						step = 0.05,
					},
				},
			},			
			active = {
				type = 'toggle',
				name = L["Active"],
				desc = L["Activate/deactivate operation."],
				order = 1,
				get  = function() return self:GetSetting("active") end,
				set  = function()
					self:ToggleSetting("active")
				end,
			},
			minimap = {
				type = 'toggle',
				name = L["Minimap Button"],
				desc = L["Show Minimap button"],
				order = 2,
				get  = function() return self:GetSetting("minimap") end,
				set  = function()
					self:ToggleSetting("minimap")
				end,
			},
			hint = {
				type = 'toggle',
				name = L["Hide Hint"],
				desc = L["Hide usage hint in tooltip."],
				order = 3,
				get  = function() return self:GetSetting("hideHint") end,
				set  = function()
					self:ToggleSetting("hideHint") 
				end,
			},
		},
	}
	
	self:SetupExpansionOptions()
	self:SetupBlacklistOptions()
end

function Options:Update()
	AceConfigReg:NotifyChange(Addon.FULLNAME)
end

function Options:SetupExpansionOptions()
	local order = 3

	local currentExpansion = GetExpansionLevel()

	for expansion = 0, currentExpansion do
		local expansionName = _G["EXPANSION_NAME"..tostring(expansion)] or "unknown"
		
		local checkbox = {
			type = 'toggle',
			name = expansionName,
			desc = string.format(L["Track nodes of expansion %s."], expansionName),
			order = order,
			get  = function() 
				return self:IsExpansionSet(expansion)
			end,
			set  = function()
				self:ToggleExpansion(expansion)
			end,
		}
		
		self.options.args.nodes.args[expansionName] = checkbox
		
		order = order + 1
	end
end

function Options:SetupBlacklistOptions()
	for profession in pairs(Addon.PROFESSIONS) do
		self:SetupBlacklistOptionsFor(profession)
	end
end

function Options:SetupBlacklistOptionsFor(profession)
	if not profession or not professionConfig[profession] then
		return
	end
	
	local conf = professionConfig[profession]
	
	self.options.args.nodes.args[conf.key] = {
		type = 'group',
		name = conf.name,
		order = 40,
		desc = conf.desc,
		args = {
			-- to be filled below
		},
	}
	
	local currentExpansion = GetExpansionLevel()

	for expansion = 0, currentExpansion do
		local order = 100*expansion
		
		local nodes = LibNTI:GetNodesForExpansion(profession, expansion)

		if next(nodes) then
			local expansionName = _G["EXPANSION_NAME"..tostring(expansion)] or "unknown"
			
			self.options.args.nodes.args[conf.key].args[expansionName] = {
				type = "header",
				name = NS:Colorize("Orange", expansionName),
				order = order,
			}
					
			table.sort(nodes, NS.compareNodes)
			
			for index, nodeId in ipairs(nodes) do
				local nodeName  = LibNTI:GetName(profession, nodeId)
				
				self.options.args.nodes.args[conf.key].args[nodeName] = {
					type = 'toggle',
					name = nodeName,
					desc = string.format(L["Track %s."], nodeName),
					order = order + index,
					get  = function() 
						return not self:IsNodeTypeBlacklisted(nodeId)
					end,
					set  = function()
						self:ToggleNodeTypeBlacklisted(nodeId)
					end,
				}
			end
		end
	end
end

-- inlet options
function Options:GetInletOptions()
	NS:ClearTable(inletOptions)

	for name, inlet in Inlets:IterateInlets() do
		if inlet:IsReady() then
			inletOptions[name] = NS:Colorize("White", name)
		else
			inletOptions[name] = NS:Colorize("GrayOut", name)
		end
	end
	
	inletOptions[Inlets.NONE] = NS:Colorize("White", L[Inlets.NONE])

	return inletOptions
end

-- algorithm options
function Options:GetAlgorithmOptions()
	NS:ClearTable(algorithmOptions)

	for name, algorithm in Algorithms:IterateAlgorithms() do
		if algorithm:IsReady() then
			algorithmOptions[name] = NS:Colorize("White", name)
		else
			algorithmOptions[name] = NS:Colorize("GrayOut", name)
		end
	end
	
	algorithmOptions[Algorithms.NONE] = NS:Colorize("White", L[Algorithms.NONE])

	return algorithmOptions
end

-- settings
function Options:GetSetting(option)
	return self.db.profile[option]
end

function Options:SetSetting(option, value)
	local current = self:GetSetting(option)

	if current == value then
		return
	end
	
	self.db.profile[option] = value

	-- fire event when setting changed
	self:SendMessage(ADDON .. "_SETTING_CHANGED", option, value, current)
end

function Options:ToggleSetting(option)
	self:SetSetting(option, not self:GetSetting(option) and true or false)
end

function Options:ToggleSettingTrueNil(option)
	self:SetSetting(option, not self:GetSetting(option) and true or nil)
end

-- node settings
function Options:IsZoneSpecific()
	return self.db.profile.nodes.zoneSpecific
end

function Options:SetZoneSpecific(active)
	active = active and true or false
	
	local current = self:IsZoneSpecific()
	
	if active ~= current then
		self.db.profile.nodes.zoneSpecific = active
	
		-- fire event when setting changed
		self:SendMessage(ADDON .. "_SETTING_CHANGED", "zoneSpecific", active, current, "nodes")
	end	
end

function Options:ToggleZoneSpecific()
	self:SetZoneSpecific(not self:IsZoneSpecific())
end

function Options:IsExpansionSet(expansion)
	return self.db.profile.nodes.expansions[expansion] and true or false
end

function Options:SetExpansion(expansion, active)
	active = active and true or false
	
	local current = self:IsExpansionSet(expansion)
	
	if active ~= current then
		self.db.profile.zones.expansions[expansion] = active
	
		-- fire event when setting changed
		self:SendMessage(ADDON .. "_SETTING_CHANGED", expansion, active, current, "nodes", "expansions")
	end	
end

function Options:ToggleExpansion(expansion)
	self:SetExpansion(expansion, not self:IsExpansionSet(expansion))
end

function Options:IsNodeTypeBlacklisted(nodeId)
	return self.db.profile.nodes.blacklist[nodeId] and true or false
end

function Options:SetNodeTypeBlacklisted(nodeId, active)
	local current = self:IsNodeTypeBlacklisted(nodeId)
	
	active = active and true or false
	
	if active == current then
		return
	end
	
	self.db.profile.nodes.blacklist[nodeId] = active
	
	self:SendMessage(ADDON .. "_SETTING_CHANGED", "blacklist", active, current, "nodes", nodeId)
end

function Options:ToggleNodeTypeBlacklisted(nodeId)
	self:SetNodeTypeBlacklisted(nodeId, not self:IsNodeTypeBlacklisted(nodeId) and true or false)
end

-- zone settings
function Options:GetZoneLevel()
	return self.db.profile.zones.level
end

function Options:SetZoneLevel(value)
	local current = self:GetZoneLevel()
	
	if value ~= current then
		self.db.profile.zones.level = value
	
		-- fire event when setting changed
		self:SendMessage(ADDON .. "_SETTING_CHANGED", "level", value, current, "zones")
	end
end

function Options:GetMultiZone()
	return self.db.profile.zones.multi
end

function Options:SetMultiZone(value)
	local current = self:GetMultiZone()
	
	if value ~= current then
		self.db.profile.zones.multi = value
	
		-- fire event when setting changed
		self:SendMessage(ADDON .. "_SETTING_CHANGED", "multi", value, current, "zones")
	end
end

function Options:IsZoneFactionSet(faction)
	return self.db.profile.zones.factions[faction] and true or false
end

function Options:SetZoneFaction(faction, active)
	active = active and true or false
	
	local current = self:IsZoneFactionSet(faction)
	
	if active ~= current then
		self.db.profile.zones.factions[faction] = active
	
		-- fire event when setting changed
		self:SendMessage(ADDON .. "_SETTING_CHANGED", faction, active, current, "zones", "factions")
	end	
end

function Options:ToggleZoneFaction(faction)
	self:SetZoneFaction(faction, not self:IsZoneFactionSet(faction))
end

function Options:IsZoneTypeSet(type)
	return self.db.profile.zones.types[type] and true or false
end

function Options:SetZoneType(type, active)
	active = active and true or false
	
	local current = self:IsZoneTypeSet(type)
	
	if active ~= current then
		self.db.profile.zones.types[type] = active
	
		-- fire event when setting changed
		self:SendMessage(ADDON .. "_SETTING_CHANGED", type, active, current, "zones", "types")
	end	
end

function Options:ToggleZoneType(type)
	self:SetZoneType(type, not self:IsZoneTypeSet(type))
end

-- outlet settings
function Options:RegisterOutletOptions(outlet)
	if not NS:IsValidOutlet(outlet) then
		return
	end
	
	local name = outlet:GetName()

	local outlet_db = self.db.profile.outlets[name]
	
	if not outlet_db then
		outlet_db = {}
		
		outlet_db.active = true
		
		self.db.profile.outlets[name] = outlet_db
	end
	
	outletOrder = outletOrder + 10
	
	local order = outletOrder
	
	-- description
	self.options.args.outlets.args[name.."-description"] = {
		type = "description",
		name = NS:Colorize("Yellow", name) .. ": " .. outlet:GetDescription(),
		order = order + 1,
	}	
	
	-- active
	local values = {}
	
	self.options.args.outlets.args[name.."-active"] = {
		type = 'select',
		name = L["Active"],
		desc = L["Activate/Deactivate outlet. If grayed out the outlet is not ready for operation. For example if an addon required for the outlet is not available."],
		order = order + 2,
		get = function()
			return self:IsOutletActive(name) and 1 or 0
		end,
		set = function()
			self:ToggleOutletActive(name)
		end,	
		values = function()
			if outlet:IsReady() then
				values[0] = NS:Colorize("Red",   L["disabled"])
				values[1] = NS:Colorize("Green", L["enabled"])
			else
				values[0] = NS:Colorize("GrayOut", L["disabled"])
				values[1] = NS:Colorize("GrayOut", L["enabled"])
			end
			
			return values
		end,
	}	
end

function Options:UnregisterOutletOptions(outlet)	
	if not NS:IsValidOutlet(outlet) then
		return
	end
	
	local name = outlet:GetName()

	-- remove outlet active checkbox to outlet page
	self.options.args.outlets.args[name.."-description"] = nil
	self.options.args.outlets.args[name.."-active"]      = nil
end

function Options:IsOutletActive(name)
	return self.db.profile.outlets[name] and self.db.profile.outlets[name].active or false
end

function Options:SetOutletActive(name, active)
	local current = self:IsOutletActive(name)
	
	active = active and true or false

	if active == current then
		return
	end
	
	self.db.profile.outlets[name].active = active
	
	self:SendMessage(ADDON .. "_SETTING_CHANGED", "outlet", active, current, name)
end

function Options:ToggleOutletActive(name)
	self:SetOutletActive(name, not self:IsOutletActive(name))
end

-- algorithm settings
function Options:RegisterAlgorithmOptions(algorithm)
	if not NS:IsValidAlgorithm(algorithm) then
		return
	end
	
	local name = algorithm:GetName()

	local algorithm_db = self.db.profile.algorithms[name]
	
	if not algorithm_db then
		algorithm_db = {}
		
		self.db.profile.algorithms[name] = algorithm_db
	end
	
	-- add algorithm to options
	self.options.args.algorithms.args[name] = algorithm:SetupOptions(algorithm_db)
end

function Options:UnregisterAlgorithmOptions(algorithm)	
	if not NS:IsValidAlgorithm(algorithm) then
		return
	end
	
	local name = algorithm:GetName()

	-- remove algorithm from options
	self.options.args.algorithms.args[name] = nil
end

-- tooltip settings
function Options:GetTooltipDetails(option)
	return self.db.profile.tooltip[option] and true or false
end

function Options:SetTooltipDetails(option, active)
	local current = self:GetTooltipDetails(option)
	
	active = active and true or false
	
	if active == current then
		return
	end
	
	self.db.profile.tooltip[option] = active
	
	self:SendMessage(ADDON .. "_SETTING_CHANGED", option, active, current, "tooltip")
end

function Options:ToggleTooltipDetails(option)
	self:SetTooltipDetails(option, not self:GetTooltipDetails(option))
end

-- test
function Options:Debug(msg)
	Addon:Debug("(Options) " .. msg)
end
