local _G = _G

-- addon name and namespace
local ADDON, NS = ...

local Addon = LibStub("AceAddon-3.0"):GetAddon(ADDON)

-- get translations
local L = LibStub:GetLibrary("AceLocale-3.0"):GetLocale(ADDON)

-- setup libs
local LibStub   = LibStub
local LDB       = LibStub:GetLibrary("LibDataBroker-1.1")
local AceTimer  = LibStub:GetLibrary("AceTimer-3.0")

-- the module name
local MODULE      = "Data Broker"
local DESCRIPTION = L["Module that provides direction arrow and force representation as DataBroker display."]

-- local functions
local cos     = math.cos
local sin     = math.sin

local GetCVar         = _G.GetCVar
local GetPlayerFacing = _G.GetPlayerFacing

local PI               = math.pi
local ONE_DIV_PI       = 1 / PI

local ARROW = "Interface\\Addons\\"..ADDON.."\\arrow.tga"

local FORCE_COLORS = {
	[0] = NS:Colorize("GrayOut", "*****"),
	[1] = NS:Colorize("Red", "*") .. NS:Colorize("GrayOut", "****"),
	[2] = NS:Colorize("Red", "*") .. NS:Colorize("Yellow", "*") .. NS:Colorize("GrayOut", "***"),
	[3] = NS:Colorize("Red", "*") .. NS:Colorize("Yellow", "**") .. NS:Colorize("GrayOut", "**"),
	[4] = NS:Colorize("Red", "*") .. NS:Colorize("Yellow", "**") .. NS:Colorize("Green", "*") .. NS:Colorize("GrayOut", "*"),
	[5] = NS:Colorize("Red", "*") .. NS:Colorize("Yellow", "**") .. NS:Colorize("Green", "**"),
}

-- aux variables
local _

-- the module
local Outlet = NS:CreateOutlet(MODULE, DESCRIPTION)

AceTimer:Embed(Outlet)

Outlet.magnitude = 0
Outlet.direction = 0
Outlet.force     = 0
Outlet.rating    = 0
Outlet.angle     = 0
Outlet.iconCoords = NS:NewTable()

local ldbName = "Magnetlet"

local ldbObj = LDB:NewDataObject(ldbName, {
	type	= "data source",
	icon	= ARROW,
	label	= ldbName,
	text	= FORCE_COLORS[0],
	OnEnter = function(self)
		Addon:CreateTooltip(self)
	end,
})
	
Outlet:SetReady(ldbObj and true or false)

-- register
NS:RegisterModule(Outlet)

-- implement functions for output
function Outlet:HandleActiveChanged()
	if not self:IsReady() then
		return
	end
	
	if self:IsActive() then
		-- setup timer if not present
		if not self.timer then
			self:ScheduleRepeatingTimer("UpdateIcon", 0.1)
		end
	else
		-- reset timer
		if self.timer then
			self:CancelTimer(self.timer)
		end
	end
end

function Outlet:Refresh()
	if not self:IsOperational() then
		return
	end
	
	local magnitude, direction, sinDir, cosDir, rating = Addon:GetForce()

	if magnitude == 0 then
		self:Clear()
		return
	end
	
	self.magnitude    = magnitude
	self.direction    = direction	
	self.rating       = rating

	-- update text
	ldbObj.text = FORCE_COLORS[self.rating]
end

function Outlet:Clear()
	if not self:IsReady() then
		return
	end

	self.magnitude = 0
	self.direction = 0	
	self.rating    = 0
	
	-- point upward
	self:SetIconTexCoord(0, 0, 0, 1, 1, 0, 1, 1)
	
	-- gray out arrow
	self:SetIconColor(0.6, 0.6, 0.6)
	
	-- reset text
	ldbObj.text = FORCE_COLORS[self.rating]
end

function Outlet:UpdateIcon()
	if not self:IsOperational() or self.magnitude == 0 then
		return
	end

	local facing = - self.direction + GetPlayerFacing()
		
	local xUL, yUL, xLL, yLL, xUR, yUR, xLR, yLR
	
	local cosDir = cos(facing)
	local sinDir = sin(facing)
	
	local gX = (sinDir + cosDir) * 0.5
	local gY = (sinDir - cosDir) * 0.5
	
	xUL = 0.5 - gX
	yUL = 0.5 + gY
	xLL = 0.5 + gY
	yLL = 0.5 + gX
	xUR = 0.5 - gY
	yUR = 0.5 - gX
	xLR = 0.5 + gX
	yLR = 0.5 - gY
	
	self:SetIconTexCoord(xUL, yUL, xLL, yLL, xUR, yUR, xLR, yLR)
	self:SetDirectionColor(facing)
end

function Outlet:SetIconTexCoord(xUL, yUL, xLL, yLL, xUR, yUR, xLR, yLR)
	local coords = NS:NewTable(xUL, yUL, xLL, yLL, xUR, yUR, xLR, yLR)
	
	NS:ReleaseTable(self.iconCoords)	
	self.iconCoords = coords
	
	ldbObj.iconCoords = self.iconCoords
end

function Outlet:GetIconTexCoord()
	return unpack(ldbObj.iconCoords)
end

-- can colors be done?
function Outlet:SetDirectionColor(angle)
	local perc = abs(PI - abs(angle)) * ONE_DIV_PI
    local g = 1
    local r = 0
	
    if perc > 0.5 then
      g = 1
      r = perc * -2 + 2
    else
      r = 1
      g = perc * 2
    end
	
    self:SetIconColor(r, g, 0)
end

function Outlet:SetIconColor(r, g, b)
	ldbObj.iconR = r
	ldbObj.iconG = g
	ldbObj.iconB = b
end

-- test
function Outlet:Debug(msg)
	Addon:Debug("(Outlet:" .. MODULE .. ") " .. msg)
end
